/****************************************************************************
*
*  wavetool.cpp - wave audio utility
*
*  Release: 0.22
*  Create:  24.04.2001
*  Update:  03.05.2001
*  (C)      Marcel Mller
*
*  This program is free software; you can redistribute it and/or
*  modify it under the terms of the GNU General Public License
*  as published by the Free Software Foundation; either version 2
*  of the License, or (at your option) any later version.
*
*  This program is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU General Public License for more details.
*
*  You should have received a copy of the GNU General Public License
*  along with this program; if not, write to the Free Software
*  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*
****************************************************************************/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <ctype.h>
#include <math.h>
#include <float.h>
#include <time.h>
#define INCL_DOSFILEMGR
#define INCL_DOSPROCESS
#include <os2.h>

#define HF_STDIN        0

#ifndef M_PI
#define M_PI            3.14159265358979323846
#endif
#ifndef M_SQRT3
#define M_SQRT3         1.73205080756887729353
#endif

#define n               (128*1024)       // blocksize (128k words)
#define vollimit        32767
#define nlclipf         (3*32767)


static HFILE      fi = HF_STDIN;
static HFILE      fo = 0;
static float      b2[n];
static float      b[n];
static short      tmp[n];
static float      kernel[n];
static long       fcnt;
static bool       eof         = false;
static long       ifsize      = 0;
static int        maxval[2]   = {INT_MIN,INT_MIN};
static int        minval[2]   = {INT_MAX,INT_MAX};
static long long  sum[2]      = {0,0};
static long long  qsum[2]     = {0,0};
static long long  Tsum[2]     = {0,0};
static long long  Tqsum[2]    = {0,0};
static double     maxavg[3]   = {-1./0,-1./0,-1./0};
static double     minavg[3]   = {+1./0,+1./0,+1./0};
static double     maxvol[3]   = {0,0};
static long       clips       = 0;

static long       histdata[2][65536];

enum initflag
{  IF_None,
   IF_Init,
   IF_End
};

static bool       autoZ       = false;
static bool       advbound    = true;
static double     subsonicf   = 15;
static bool       truncfile    = false;
static bool       rawfile     = false;
static bool       frameinfo   = false;
static double     framesize   = 0;
static double     fframesize  = .5;
static double     newvol      = 0;
static double     thres       = 1;
static bool       noise       = true;
static bool       monovol     = true;
static bool       interactive = false;
static enum
{  noClipCheck,
   ClipCheck,
   AutoClip
}                 clip        = AutoClip;
static bool       nlclip      = false;
static bool       clipreport  = false;
static const char*histf       = NULL;
static const char*kernelf     = NULL;
static int        histmask    = 0xFFFFFF80;
static const char*fname       = NULL;
static const char*ofname      = NULL;


template<class T>
static inline T min(const T& v1, const T& v2)
{  return v1 < v2 ? v1 : v2;
}
template<class T>
static inline T max(const T& v1, const T& v2)
{  return v1 > v2 ? v1 : v2;
}
template<class T>
static inline void storemin(T& where, const T what)
{  if (what < where)
      where = what;
}
template<class T>
static inline void storemax(T& where, const T what)
{  if (what > where)
      where = what;
}

static inline double todB(double v)
{  return 20 * log10(v);
}

static void error(unsigned long r, const char* fmt, int ret = 1)
{  DosClose(fi);
   DosClose(fo);
   fprintf(stderr, fmt, r);
   exit(ret);
}
static inline void errchk(unsigned long r, const char* fmt, int ret = 1)
{  if (r)
      error(r, fmt, ret);
}

/* assembler funcs */
extern "C" {
int has_MMX_nasm(void);
int has_SIMD_nasm(void);
int has_SIMD2_nasm(void);
int has_3DNow_nasm(void);
void scalar_3DNow_nasm(float* dp, const float* s1, const float* s2, unsigned int len);
}

void scalar(float* dp, const float* sp, const float* kp, unsigned int len)
{  double sum[4] = {0,0,0,0};
   while (len--)
   {  sum[0] += sp[0] * kp[0];
      sum[1] += sp[1] * kp[0];
      sum[2] += sp[2] * kp[0];
      sum[3] += sp[3] * kp[0];
      sp += 2;
      ++kp;
   }
   dp[0] = sum[0];
   dp[1] = sum[1];
   dp[2] = sum[2];
   dp[3] = sum[3];
}

static short      (*rescanf)(double f);
static void       (*filterf)(const float* sp, unsigned long len, initflag iflg) = NULL;
static void       (*scalarf)(float* dp, const float* s1, const float* s2, unsigned int len) = scalar;


static unsigned long readwav(unsigned long l)
{  unsigned long rq = l;
   if (rq > n/2)
      rq = n/2;
   rq <<= 2;
   unsigned long len = rq;
   short* dp = tmp;
   unsigned long x;
   do
   {  errchk(DosRead(fi, dp, len, &x), "Failed to read wave data - Error %lu.\n");
      len -= x;
      dp += x;
      if (x == 0)
      {  fprintf(stderr, "Failed to read wave data - EOF reached. Read %lu, Requested %lu.\n", rq - len, l << 2);
         eof = true;
         break;
      }
   } while (len);
   return (rq - len) >> 2;
}

static void tofloat(float* dp, const short* sp, unsigned long l)
{  while (l--)
   {  *dp++ = *sp++;
      *dp++ = *sp++;
}  }

static void toint(short* dp, const float* sp, unsigned long l)
{  while (l--)
   {  *dp++ = (*rescanf)(*sp++);
      *dp++ = (*rescanf)(*sp++);
}  }

static void write(const void* data, unsigned long l)
{  unsigned long r2;
   errchk(DosWrite(fo, (PVOID)data, l, &r2), "Failed to write wave data - Error %lu.\n");
   if (r2 != l)
      error(0, "Failed to write wave data - Disk full.\n");
}

static inline void writewav(unsigned long l)
{  write(tmp, 4*l);
}

static void docalc(double* avg, double* vol, const long long* sum, const long long* qsum, long l)
{  avg[0] = (double)sum[0]/l;
   avg[1] = (double)sum[1]/l;
   avg[2] = (double)(sum[0]+sum[1])/(l<<1);
   vol[0] = sqrt((double)qsum[0]/l-avg[0]*avg[0]);
   vol[1] = sqrt((double)qsum[1]/l-avg[1]*avg[1]);
   vol[2] = sqrt((double)(qsum[0]+qsum[1])/(2*l)-avg[2]*avg[2]);
}

static void nextframe(long l)
{  // evaluate frame
   if (l == ifsize)
   {  double avg[3], vol[3];
      docalc(avg, vol, sum, qsum, l);
      if (frameinfo)
         printf("Frameinfo:   %8.2f  %8.2f  %8.2f\n"
                "             %8.1f  %8.1f  %8.1f\n"
          , avg[0], avg[1], avg[2],
            vol[0], vol[1], vol[2]);
      storemax(maxvol[0], vol[0]);
      storemax(maxvol[1], vol[1]);
      storemax(maxvol[2], vol[2]);
      storemax(maxavg[0], avg[0]);
      storemax(maxavg[1], avg[1]);
      storemax(maxavg[2], avg[2]);
      storemin(minavg[0], avg[0]);
      storemin(minavg[1], avg[1]);
      storemin(minavg[2], avg[2]);
   }
   // add to totals
   Tsum[0] += sum[0];
   Tsum[1] += sum[1];
   Tqsum[0] += qsum[0];
   Tqsum[1] += qsum[1];
   // reset
   sum[0] = 0;
   sum[1] = 0;
   qsum[0] = 0;
   qsum[1] = 0;
   fcnt = ifsize;
}

static void scanbuf(const short* sp, unsigned long len)
{  while (len--)
   {  register int val = *sp++;
      storemax(maxval[0], val);
      storemin(minval[0], val);
      sum[0] += val;
      qsum[0] += val * val;
      bool clip = clipreport && (val == -32768 || val == 32767);
      val = *sp++;
      storemax(maxval[1], val);
      storemin(minval[1], val);
      sum[1] += val;
      qsum[1] += val * val;
      if (clipreport && (clip || val == -32768 || val == 32767))
      {  unsigned long pos;
         DosSetFilePtr(fi, 0, FILE_CURRENT, &pos);
         printf("possible clip: offset %08X value %04.4X %04.4X\n"
          , pos - (len+1 <<2), sp[-2], val);
      }
      // check frame end
      if (!--fcnt)
         nextframe(ifsize);
   }
}

static void createhist(unsigned long len)
{  register short* sp = tmp;
   while (len--)
   {  ++histdata[0][histmask & *sp++ + 32768];
      ++histdata[1][histmask & *sp++ + 32768];
   }
}

// rescan functions
static short RSFnonoise(double ival)
{  int val = (int)floor(ival + .5);
   if (val >= 32768)
   {  printf("\t+clip %i\r", ++clips);
      val = 32767;
   } else if (val < -32768)
   {  printf("\t-clip %i\r", ++clips);
      val = -32768;
   }
   return (short)val;
}

static short RSFstd(double ival)
{  int val = (int)floor(ival + rand() / (RAND_MAX + 1.));
   if (val >= 32768)
   {  printf("\t+clip %i\r", ++clips);
      val = 32767;
   } else if (val < -32768)
   {  printf("\t-clip %i\r", ++clips);
      val = -32768;
   }
   return (short)val;
}

static short RSFsoftnonoise(double val)
{  val *= M_PI/nlclipf;
   val = nlclipf/12./M_PI *
    ( 4.*atan(val)
    - 2.*atan(M_SQRT3-2.*val)
    + 2.*atan(M_SQRT3+2.*val)
    - M_SQRT3 * log(1.-M_SQRT3*val+val*val)
    + M_SQRT3 * log(1.+M_SQRT3*val+val*val)
    );
   return (short)floor(val + .5);
}

static short RSFsoft(double val)
{  val *= M_PI/nlclipf;
   val = nlclipf/12./M_PI *
    ( 4.*atan(val)
    - 2.*atan(M_SQRT3-2.*val)
    + 2.*atan(M_SQRT3+2.*val)
    - M_SQRT3 * log(1.-M_SQRT3*val+val*val)
    + M_SQRT3 * log(1.+M_SQRT3*val+val*val)
    );
   return (short)floor(val + rand() / (RAND_MAX + 1.));
}

// operators
static void setZ(unsigned long len, double* offset)
{  register float* sp = b;
   while (len--)
   {  sp[0] += offset[0];
      sp[1] += offset[1];
      sp += 2;
   }
}

static void adjvol(unsigned long len, double* scale)
{  register float* sp = b;
   while (len--)
   {  sp[0] *= scale[0];
      sp[1] *= scale[1];
      sp += 2;
   }
}

static unsigned long filterOP(unsigned long len)
{  static bool init = false;
   float* sp;
   if (!init)
   {  // initial call
      init = true;
      //printf("initial %lu %f,%f\n", len, avg[0],avg[1]);
      if (advbound)
      {  //printf("advanced %i\n",autoZ);
         double avg[2] = {0,0};
         long l = ifsize;
         sp = b;
         while (l--)
         {  avg[0] += sp[0];
            avg[1] += sp[1];
            sp += 2;
         }
         avg[0] /= ifsize;
         avg[1] /= ifsize;
         float* dp = b - 2*ifsize;
         l = ifsize;
         while (l--)
         {  dp[0] = avg[0];
            dp[1] = avg[1];
            dp += 2;
         }
      } else
         memset(b - 2*ifsize, 0, 2*ifsize * sizeof *b);
      len -= ifsize;
      (*filterf)(sp = b, len, IF_Init);
   } else if (!len)
   {  // final call
      //printf("final %lu\n", len);
      if (advbound)
      {  double avg[2] = {0,0};
         len = ifsize;
         sp = b - 2*ifsize;
         while (len--)  // scan last half frame
         {  avg[0] += sp[0];
            avg[1] += sp[1];
            sp += 2;
         }
         //printf("avx %f,%f\n",avg[0],avg[1]);
         avg[0] /= ifsize;
         avg[1] /= ifsize;
         float* dp = b;
         len = ifsize;
         while (len--)
         {  dp[0] = avg[0];
            dp[1] = avg[1];
            dp += 2;
         }
      } else
         memset(b, 0, 2*ifsize * sizeof *b);
      len = ifsize;
      (*filterf)(sp = b - 2*ifsize, len, IF_End);
   } else
      (*filterf)(sp = b - 2*ifsize, len, IF_None);
   //printf("avg %lf,%lf %p,%p\n", avg[0],avg[1], sp,dp);
   toint(tmp, b2, len);
   // save last samples
   memcpy(b - 4*ifsize, sp + 2*len - 2*ifsize, 4*ifsize * sizeof *b);
   return len;
}

static void FIRmovingavg(const float* sp, unsigned long len, initflag iflg)
{  static double avg[2] = {0,0};
   long offset = ifsize << 1;
   float* dp = b2;
   if (iflg & IF_Init)
   {  // initialize average
      sp -= offset;
      int l = offset;
      while (l--)
      {  avg[0] += sp[0];
         avg[1] += sp[1];
         sp += 2;
      }
      sp -= offset;
   }
   double divf = offset +1;
   while (len--)
   {  avg[0] += sp[offset];
      dp[0] = sp[0] - avg[0]/divf;
      avg[0] -= sp[-offset];
      avg[1] += sp[offset+1];
      dp[1] = sp[1] - avg[1]/divf;
      avg[1] -= sp[-offset+1];
      sp += 2;
      dp += 2;
}  }

static void FIRgeneric(const float* sp, unsigned long len, initflag iflg)
{  float* dp = b2;
   sp -= 2*ifsize;
   // convolution
   while (len)
   {  (*scalarf)(dp, sp, kernel, 2*ifsize +1);
      dp += 4;
      sp += 4;
      if (len == 1)
         break;
      len -= 2;
}  }

static int readkernel(const char* file)
{  FILE* f = fopen(file, "r");
   if (f == NULL)
   {  fprintf(stderr, "failed to open kernel file %s.\n", file);
      return 3;
   }
   float* dp = kernel;
   while(fscanf(f, "%f", dp) == 1)
      if (++dp == *(&kernel+1))
      {  fprintf(stderr, "kernel file exceeds maximum filter order %u.\n", n);
         fclose(f);
         return 3;
      }
   fclose(f);
   ifsize = dp - kernel;
   if (ifsize == 0)
   {  fprintf(stderr, "kernel file %s does not contain valid data.\n", file);
      return 3;
   }
   if ((ifsize & 1) == 0)
   {  fputs("Warning: filter kernel has even length.\n"
            " Using linear phase transformation (time shift = +1/2 sample).", stderr);
      *dp = 0;
      ++ifsize;
   }
   return 0;
}

static void printprogress(double pct)
{  static time_t t = (time_t)-1;
   if (pct == 0.)
   {  t = clock();
      fputs("0.0 %\r", stderr);
      return;
   } else if (pct == 1.)
   {  double eta = (clock()-t)/CLOCKS_PER_SEC;
      int m = (int)(eta/60);
      fprintf(stderr, "100.0 %% - Total: %i:%04.1f\n", m, eta-60*m);
      return;
   }
   double eta = (1-pct) / pct * (clock()-t)/CLOCKS_PER_SEC;
   int m = (int)(eta/60);
   fprintf(stderr, "%.1f %% - ETA: %i:%04.1f \r", 100.*pct, m, eta-60*m);
   fflush(stderr);
}

int parseoption(const char* cp)
{  switch (cp[0])
   {default:
      fprintf(stderr, "illegal option %s.\n", cp);
      return 4;
    case 'o':
      ofname = cp +1;
      break;
    case 't':
      truncfile = true;
      break;
    case 'z':
      switch (cp[1])
      {default:
         fprintf(stderr, "illegal option %s.\n", cp);
         return 4;
       case 'm':
         filterf = FIRmovingavg;
         break;
       case 'f':
         filterf = FIRgeneric;
         if (cp[2] && (sscanf(cp +2, "%f", &subsonicf) != 1 || subsonicf <= 0))
         {  fprintf(stderr, "illegal subsonic frequency %s.\n", cp +2);
            return 3;
         }
         break;
       case 0:
         autoZ = true;
      }
      break;
    case 'a':
      advbound = false;
      break;
    case 'r':
      rawfile = true;
      break;
    case 'f':
      framesize = 1;
      sscanf(cp +1, "%lf", &framesize);
      break;
    case 'F':
      sscanf(cp +1, "%lf", &fframesize);
      break;
    case 'v':
      frameinfo = true;
      break;
    case 'l':
      sscanf(cp +1, "%lf", &newvol);
      if (!newvol)
      {  fprintf(stderr, "illegal volume %s.\n", cp +2);
         return 4;
      }
      break;
    case 'p':
      thres = 0;
      sscanf(cp +1, "%lf", &thres);
      if (!thres)
      {  fprintf(stderr, "illegal threshold %s.\n", cp +2);
         return 4;
      }
      break;
    case 'n':
      noise = false;
      break;
    case 's':
      monovol = false;
      break;
    case 'i':
      interactive = true;
      break;
    case 'c':
      clip = noClipCheck;
      if (cp[1] == 'n')
         nlclip = true;
      break;
    case 'C':
      clipreport = true;
      break;
    case 'h':
      histf = cp +1;
      break;
    case 'k':
      filterf = FIRgeneric;
      subsonicf = 0;
      return readkernel(cp+2);
    case 'K':
      kernelf = cp +1;
      break;
   }
   return 0;
}

int main(int argc, char** argv)
{  if (argc < 2)
   {  fputs("Usage: WAVEINFO <filename> [<options>]\n"
            "options: /ofile output file\n"
            "                By default the operation is in place.\n"
            "         /z   auto correct zero line\n"
            "         /zm  moving average subsonic\n"
            "         /zf# FIR subsonic (uses 3Dnow, slow either)\n"
            "            # transition frequency [Hz], default: 15\n"
            "         /a   no advanced boundary conditions for FIR filters\n"
            "         /r   <filename> is raw file\n"
            "         /t   truncate additional data\n"
            "         /f#  calculate frame peak volume\n"
            "           #  frame length [s], default: 1\n"
            "         /F#  Frame length for FIR filter kernels, default 0.5s\n"
            "         /v   view detailed frame infos\n"
            "         /l#  change volume to ...\n"
            "           #  new volume [dB|scalefactor]\n"
            "         /p#  change volume only if it differs by factor #\n"
            "         /n   no noise for requantization (dithering)\n"
            "         /s   scale volume per channel (stereo)\n"
            "         /i   interactive mode\n"
            "         /c   no clipping check\n"
            "         /cn  soft (nonlinear) clipping\n"
            "         /C   possible clip report\n"
            "         /hfile value histogram -> file\n"
            "         /kfile read filter kernel from file\n"
            "         /Kfile write filter kernel -> file\n"
       , stderr);
      /*printf("%p %p\n",scalar_3DNow_nasm);*/
      return 6;
   }
   for (int i = 1; i < argc; ++i)
   {  char* cp = argv[i];
      if (*cp == '/')
      {  int r = parseoption(cp+1);
         if (r)
            return r;
      } else if (fname)
      {  fputs("Syntax error in command line.\n", stderr);
         return 5;
      } else
         fname = cp;
   }
   if (!fname)
   {  fputs("Filename required.\n", stderr);
      return 5;
   }
   truncfile &= !rawfile;

   // CPU dependent implementations
   if (has_3DNow_nasm())
   {  scalarf = scalar_3DNow_nasm;
      fputs("3DNow detected.\n", stderr);
   }

   struct
   {  char  RIFF[4];
      long  BlkLen;
      char  WAVE[8];
      long  WAVLen;
      short x1;
      short Channels;
      long  SampRate;
      long  x2;
      short BpSamp;
      short Bits;
      char  data[4];
      long  DataLen;
   } Header;

   ULONG r, r2;

   // open file
   if (fname[0] != '-' || fname[1] != 0)
   {  r = DosOpen((PUCHAR)fname, &fi, &r2, 0, 0, OPEN_ACTION_FAIL_IF_NEW|OPEN_ACTION_OPEN_IF_EXISTS,
       OPEN_FLAGS_SEQUENTIAL|OPEN_SHARE_DENYNONE|OPEN_ACCESS_READONLY, NULL);
      if (r)
      {  fprintf(stderr, "Cannot open %s - Error %lu.\n", fname, r);
         return 4;
   }  }

   if (!rawfile)
   {  // WAV
      errchk(DosRead(fi, &Header, sizeof Header, &r2), "Failed to read header - Error %lu.\n", 3);
      if (r2 != sizeof Header)
         error(r2, "Failed to read header - Unexpected end of file after %lu bytes.\n", 3);

      if (  memcmp(Header.RIFF, "RIFF", sizeof Header.RIFF)
         || memcmp(Header.WAVE, "WAVEfmt ", sizeof Header.WAVE)
         || Header.WAVLen != 16
         || Header.BpSamp != 4
         || Header.Bits != 16
         || memcmp(Header.data, "data", sizeof Header.data) )
      {  DosClose(fi);
         fputs("Unrecognized file format.", stderr);
         return 2;
      }
   } else if (fname[0] != '-' || fname[1] != 0)
   {  // rawfile
      FILESTATUS3 fst;
      errchk(DosQueryFileInfo(fi, FIL_STANDARD, &fst, sizeof fst), "Failed to query file info - Error %lu.\n");

      if (fst.cbFile & 3)
      {  DosClose(fi);
         fprintf(stderr, "Wave data size does not match - File: %lu.\n", fst.cbFile);
         return 2;
      }

      Header.DataLen = fst.cbFile;
      Header.SampRate = 44100;
   } else
   {  // raw streaming
      Header.DataLen = LONG_MAX;
      Header.SampRate = 44100;
   }

   // calc ifsize
   if (framesize != 0)
   {  ifsize = static_cast<int>(framesize * Header.SampRate);
      framesize = 0;
   }
   if (ifsize > Header.DataLen >> 2)
   {  fprintf(stderr, "Warning: frame length (%li) longer than file (%li)\n", ifsize, Header.DataLen >> 2);
      ifsize = Header.DataLen >> 2;
   } else if (ifsize == 0)
      ifsize = Header.DataLen >> 2;
    else
      printf("frame length = %li samples\n", ifsize);
   if (filterf && ifsize >= n)
   {  DosClose(fi);
      fprintf(stderr, "Cannot handle FIR filter with framesize of %li.\n", ifsize);
      return 1;
   }

   putchar('\n');

   unsigned long len = Header.DataLen;
   unsigned long offset = 0;
   unsigned long l, ls;

   // scan wav file
   fcnt = ifsize;
   ls = len >> 2;
   printprogress(0);
   do
   {  l = readwav(ls);
      scanbuf(tmp, l);
      if (histf)
         createhist(l);
      printprogress(4.*(offset += l)/len);
   } while ((ls -= l) && !eof);
   len -= ls << 2;

   // last frame
   nextframe(ifsize - fcnt);

   // display results
   _uldiv_t lsr = _uldiv(l, Header.SampRate);
   lsr.rem = lsr.rem * 100 / Header.SampRate;
   _uldiv_t lms = _uldiv(lsr.quot, 60);
   _uldiv_t lhm = _uldiv(lms.quot, 60);
   l = len >> 2;
   double avg[3], vol[3];
   docalc(avg, vol, Tsum, Tqsum, l);

   printf(
    "Channel:          left     right      both\n"
    "\n"
    "Minimum:        %6i    %6i    %6i\n"
    "Maximum:        %6i    %6i    %6i\n"
    "Average/DC:   %8.2f  %8.2f  %8.2f\n"
    //"Avg.Vol.:     %8.1f  %8.1f  %8.1f\n"
    "Avg.Vol. [dB]:%8.1f  %8.1f  %8.1f\n",
    minval[0], minval[1], minval[0]<minval[1]?minval[0]:minval[1],
    maxval[0], maxval[1], maxval[0]>maxval[1]?maxval[0]:maxval[1],
    avg[0], avg[1], avg[2],
    //vol[0], vol[1], vol[2],
    todB(vol[0]/32768*M_SQRT2), todB(vol[1]/32768*M_SQRT2), todB(vol[2]/32768*M_SQRT2)
   );
   if (ifsize)
      printf(//"Max.vol.:     %8.1f  %8.1f  %8.1f\n"
             "Max.vol. [dB]:%8.1f  %8.1f  %8.1f\n"
             "Min.avg.:     %8.2f  %8.2f  %8.2f\n"
             "Max.avg.:     %8.2f  %8.2f  %8.2f\n"
       //, maxvol[0], maxvol[1], maxvol[2]
       , todB(maxvol[0]/32768*M_SQRT2), todB(maxvol[1]/32768*M_SQRT2), todB(maxvol[2]/32768*M_SQRT2)
       , minavg[0], minavg[1], minavg[2], maxavg[0], maxavg[1], maxavg[2]);
    else
   {  maxvol[0] = max(-minval[0], maxval[0]) / M_SQRT2;
      maxvol[1] = max(-minval[1], maxval[1]) / M_SQRT2;
   }
   printf(
    "\nTotal Length:                  %2u:%02u:%02u.%02u\n\n",
    lhm.quot, lhm.rem, lms.rem, lsr.rem
   );

   // histogramm?
   if (histf)
   {  puts("Writing histogramm data...\n\n");
      // default file name
      if (*histf == 0)
         histf = strcat(strcpy((char*)b, fname), ".hst");
      FILE* hf = fopen(histf, "wt");
      if (hf == NULL)
         fprintf(stderr, "Error: failed to open histogram output file %s.\n", histf);
       else
      {  fputs(" Value\t     Left\t Left %\t    Right\tRight %\n", hf);
         double l = len >> 2;
         for(int i = 0; i < 65536; i -= histmask)
         {  fprintf(hf, "\n%6i", i - 32768);
            if (histdata[0][i])
               fprintf(hf, "\t%9lu\t%7.4f", histdata[0][i], histdata[0][i]/l);
             else if (histdata[1][i])
               fputs("\t\t\t", hf);
            if (histdata[1][i])
               fprintf(hf, "\t%9lu\t%7.4f", histdata[1][i], histdata[1][i]/l);
         }
         fclose(hf);
   }  }

   // interactive?
   if (interactive)
   {  puts("Enter options or hit RETURN to continue.");
      for(;;)
      {  char line[1000];
         line[0] = 0;
         gets(line);
         if (line[0] == 0)
            break;
         parseoption(line);
      }
      // recalc ifsize
      if (framesize != 0)
         ifsize = static_cast<int>(framesize * Header.SampRate);
      if (ifsize > Header.DataLen >> 2)
      {  fprintf(stderr, "Warning: frame length (%li) longer than file (%li)\n", ifsize, Header.DataLen >> 2);
         ifsize = Header.DataLen >> 2;
      } else if (ifsize == 0)
         ifsize = Header.DataLen >> 2;
      if (filterf && ifsize >= n/2)
      {  DosClose(fi);
         fprintf(stderr, "Cannot handle FIR with framesize of %li.\n", ifsize);
         return 1;
      }
   }

   // select rescan function
   static short (*const RSFtable[2][2])(double) = {{RSFnonoise,RSFsoftnonoise},{RSFstd,RSFsoft}};
   rescanf = RSFtable[noise][nlclip];

   // adjust volume
   double volscale[2];
   if (newvol <= 0)
   {  // in dB
      double nv = 32768/M_SQRT2 * pow(10, newvol/20);
      if (monovol)
         volscale[1] = volscale[0] = nv / maxvol[2];
        else
      {  volscale[0] = nv / maxvol[0];
         volscale[1] = nv / maxvol[1];
      }
   } else
      // scalefactor
      volscale[0] = volscale[1] = newvol;
   if ( volscale[0] >= 1/thres && volscale[0] <= thres
       && volscale[1] >= 1/thres && volscale[1] <= thres )
      newvol = 0;
   if (newvol)
      printf("scale volume: %8.2f  %8.2f\n", volscale[0], volscale[1]);
    else
      volscale[0] = volscale[1] = 1;

   // ifsize for FIR filter
   ifsize = (static_cast<int>(fframesize * Header.SampRate) -1) >> 1;
   //printf("ifsize: %li\n", ifsize);

   // operate
   if (autoZ || filterf || newvol)
   {  double newZ[2] = {0,0};
      if (autoZ)
      {  newZ[0] = -avg[0] * volscale[0];
         newZ[1] = -avg[1] * volscale[1];
         printf("adjust baseline:%6.1f    %6.1f\n", newZ[0], newZ[1]);
      }
      // check volume limits
      if (clip != noClipCheck)
      {  double limit = vollimit;
         storemax(limit, -minval[0]*volscale[0]-newZ[0]);
         storemax(limit, -minval[1]*volscale[1]-newZ[1]);
         storemax(limit, maxval[0]*volscale[0]+newZ[0]);
         storemax(limit, maxval[1]*volscale[1]+newZ[1]);
         if (limit > vollimit)
         {  fprintf(stderr, "Warning: Operation causes clipping. (maximum: %.0f)\n", limit);
            if (newvol && clip == AutoClip)
            {  volscale[0] *= vollimit/limit;
               volscale[1] *= vollimit/limit;
               printf("reduced scale:%8.2f  %8.2f\n", volscale[0], volscale[1]);
            } else
            {  DosClose(fi);
               return 1;
      }  }  }

      if (filterf == FIRmovingavg)
         printf("apply moving average filter with filter order %li\n", 2*ifsize);

      if (filterf == FIRgeneric && subsonicf != 0)
      {  // init filter kernel with hamming window
         double tmp = 2 * M_PI * subsonicf / Header.SampRate;
         for (int i = 0; i <= 2*ifsize; ++i)
         {  double j = i - ifsize + 1E-20;
            kernel[i] = -sin(tmp * j) / j / M_PI * (.54 - .46 * cos(i*M_PI/ifsize));
         }
         // spectral inversion
         kernel[ifsize+1] += 1;
         if (kernelf)
         {  puts("Writing filter kernel data...\n\n");
            // default file name
            if (*kernelf == 0)
               kernelf = strcat(strcpy((char*)b, fname), ".krnl");
            FILE* hf = fopen(kernelf, "wt");
            if (hf == NULL)
               fprintf(stderr, "Error: failed to open filter kernel output file %s.\n", kernelf);
             else
            {  for (int i = 0; i <= 2*ifsize; ++i)
                  fprintf(hf, "%12.8f\n", kernel[i]);
               fclose(hf);
         }  }
         printf("apply FIR filter with filter order %li\n", 2*ifsize+1);
         DosSetPriority(PRTYS_THREAD, PRTYC_IDLETIME, 20, 0);
      }

      ls = len >> 2;

      // open destination
      if (ofname == NULL)
         ofname = fname;
      r = DosOpen((PUCHAR)ofname, &fo, &r2, Header.DataLen + sizeof(Header), 0,
       OPEN_ACTION_CREATE_IF_NEW|OPEN_ACTION_OPEN_IF_EXISTS,
       OPEN_FLAGS_SEQUENTIAL|OPEN_SHARE_DENYWRITE|OPEN_ACCESS_WRITEONLY, NULL);
      if (r)
      {  fprintf(stderr, "Cannot open %s for writing - Error %lu.\n", ofname, r);
         return 4;
      }

      if (truncfile)
         // update wave header
         Header.BlkLen = Header.DataLen + sizeof(Header) -8;
      // write wave header
      write(&Header, sizeof Header);

      errchk(DosSetFilePtr(fi, rawfile ? 0 : sizeof(Header), FILE_BEGIN, &r2), "Failed to set file pointer - Error %lu.\n");
      printprogress(0);
      offset = 0;
      do
      {  ls -= l = readwav(ls);
         tofloat(b, tmp, l);
         if (newvol)
            adjvol(l, volscale);
         if (autoZ)
            setZ(l, newZ);
         if (filterf)
            l = filterOP(l);
          else
            toint(tmp, b, l);
         writewav(l);
         printprogress(4.*(offset += l)/len);
      } while (ls);
      if (filterf)
      {  l = filterOP(0);
         writewav(l);
         printprogress(4.*(offset += l)/len);
      }
   }

   if (truncfile && fname != ofname)
   {  puts("\nTruncate additional data...");
      errchk(DosSetFileSize(fo, Header.DataLen + sizeof(Header)), "Failed to truncate file - Error %lu.\n");
   }

   DosClose(fi);
   DosClose(fo);

   return 0;
}

